// Data.java
import java.util.*;

class Data {
  byte dia, mes;
  short ano;

  public Data() {};

  public Data (byte dia, byte mes, short ano) {
     this.dia = dia;
     this.mes = mes;
     this.ano = ano;
  }

  public Data (Data d) {
     dia = d.dia;
     mes = d.mes;
     ano = d.ano;
  }

  public Data (String d) {
     byte i = 0, j = 0;
     byte tam = (byte) d.length();
     if (tam < 8 || tam > 10) {
       dia = -1;
       return;
     } else {
        if (tam == 9) {
           if (d.charAt(1) == '/') {
              i = 1; 
           } else {
		  j = 1;
           }
        } else {
          if (tam == 8) {
            i = 1;
            j = 1;
          }
        }
     }
     dia = (byte) Integer.parseInt(d.substring(0,2-i));
     mes = (byte) Integer.parseInt(d.substring(3-i,5-i-j));
     ano = (short) Integer.parseInt(d.substring(6-i-j,10-i-j));
  }
  
  public boolean bissexto() {
     return (ano - 2000) % 4 == 0;  // aproximacao
  }

  public byte dia () {
     return dia;
  }

  public byte mes () {
     return mes;
  }

  public short ano () {
     return ano;
  }

  public boolean valida() {
     return !(
       (dia < 1 || dia > 31 || mes < 1 || 
          mes > 12 || ano < 1 || ano > 9999) ||    
       (dia == 29 && mes == 2 && !bissexto()) || 
       (dia == 30 && mes == 2) || 
      (dia == 31 && (mes == 2 || mes == 4 || mes == 6 || 
          mes == 9 || mes == 11))  );
  }
  
  public void modifica (Data d) { 
     dia = d.dia;
     mes = d.mes;
     ano = d.ano;
  }

  public void modifica (byte d, byte m, short a) { 
     dia = d;
     mes = m;
     ano = a;
  }
  
  public boolean menorDiaMes (Data d) {
    if (mes < d.mes) {
      return true;
    } else {
       if (mes > d.mes) {
          return false;
       } else {
          if (dia < d.dia) {
             return true;
          } else {
             return false;
          }
       }
    }
  }

  public boolean igualDiaMes (Data d) {
    return dia == d.dia && mes == d.mes;
  }

  public boolean igual (Data d) {
    return dia == d.dia && mes == d.mes && ano == d.ano;
  }

  public boolean menor (Data d) {
    if (ano < d.ano) {
       return true;
    } else {
       if (ano > d.ano) {
          return false;
       } else {
          return menorDiaMes (d);
       }
    }
  }

  short diasAno () {
    short soma = 0;
    switch (mes - 1) {
      case 11: soma += 30;
      case 10: soma += 31;
      case  9: soma += 30;
      case  8: soma += 31;
      case  7: soma += 31;
      case  6: soma += 30;
      case  5: soma += 31;
      case  4: soma += 30;
      case  3: soma += 31;
      case  2: if (bissexto()) {
                 soma += 29;
               } else {
                 soma += 28;
               }
      case  1: soma += 31;
      case  0:
      default: 
    }
    soma += dia;
    return soma;
  }

  long converteDias () {
    return (long) ((ano - 1) * 365.25 + diasAno()); //aproximacao
  }

  public Data incrementa() {
    switch (mes) {
      case 12: if (dia == 31) {
                 dia = 0;
                 mes = 1;
                 ano++;
               }
               break;
      case 11: 
      case  9: 
      case  6: 
      case  4: if (dia == 30) {
                 dia = 0;
                 mes++;
               }
               break;
      case 10:
      case 8:
      case 7:
      case 5:
      case 3:
      case 1:  if (dia == 31) {
                 dia = 0;
                 mes++;
               }
               break;
      case  2: if((dia == 28 && !bissexto())||dia == 29) {
                 dia = 0;
                 mes++;
               }
    }
    dia++;
    return this;
  }

  public long diferencaDias (Data d) {
    return Math.abs (converteDias() - d.converteDias());
  }

  public short diferencaAnos (Data d) {
    return (short) (diferencaDias(d)/365.25);
  }  

  public void leConsole() {
     Data d = null;
     boolean continua = true;
     while (continua) {
       continua = false;
       System.out.print (
          "Entre com a data no formato dd/mm/aaaa: ");
       d = new Data(Console.readString());
       if (!d.valida()) {
         System.out.println ("Erro! Tente novamente!!!");
         continua = true;
       }
     }
     modifica(d);
  }

  public void leRelogio() {
     Date d = new Date();
     dia = (byte) d.getDate();
     mes = (byte) (d.getMonth() + 1);
     ano = (short) (1900 + d.getYear());
  }

  public void imprimeFormatoIngles() {
     System.out.println (mes + "/" + dia + "/" + ano);
  }

  public String toString() {
     return dia + "/" + mes + "/" + ano;
  }

}


/* 
**  Pessoa.java 
*/

interface PessoaInter {
   String retornaNome(); 
   Data retornaAniv();
   String toString();
}

class Pessoa implements PessoaInter {
   protected String nome;
   protected Data aniv;

   public static class ComparaNome implements Comparator {
      public int compare (Object o1, Object o2) {
        PessoaInter p1 = (PessoaInter) o1;
        PessoaInter p2 = (PessoaInter) o2;
        String n1 = p1.retornaNome();
        String n2 = p2.retornaNome();
        return n1.compareToIgnoreCase(n2);
      }
   }

   public static class ComparaAniv implements Comparator {
      public int compare (Object o1, Object o2){
        PessoaInter p1 = (PessoaInter) o1;
        PessoaInter p2 = (PessoaInter) o2;
        Data d1 = p1.retornaAniv();
        Data d2 = p2.retornaAniv();
        if (d1.igual(d2)) return 0;
        if (d1.menor(d2)) return -1;
        return 1;
      }
   }

   public Pessoa (){}

   public Pessoa (String n, Data d) {
	nome = new String (n);
      aniv = new Data(d);
   }

   public Pessoa (Pessoa p) {
	nome = new String (p.nome);
      aniv = new Data(p.aniv);
   }

   public String retornaNome() {
      return new String(nome);
   }

   public Data retornaAniv() {
      return new Data(aniv);
   }

   public String toString() {
     return "Nome: " + nome + "\n" +
            "Aniversario: " + aniv;
   }

   public void leConsole() {
     System.out.print("Nome: ");
     nome = Console.readString();
     System.out.print("Aniversario: ");
     aniv = new Data();
     aniv.leConsole();
   }

}

/* 
**  MembroUniv.java 
**  Modificaes: 
**    - uso de interface 
*/

interface MUInter extends PessoaInter {
	int retornaMatr();
	Data retornaIngresso();
}

abstract class MembroUniv extends Pessoa implements MUInter {
   private static int matrAtual = 0;
   protected int matricula;
   protected Data ingresso;
  
   private static int MenuInclusao() {
     for (;;) {
       System.out.println("\nMembro da Universidade");
       System.out.println("  0 - Sair ");
       System.out.println("  1 - Assalariado");
       System.out.println("  2 - Estudante");
       System.out.println("  3 - AssalariadoEstudante");
       System.out.print("Escolha Tipo de Novo Produto:");
       int opcao = Console.readInteger();
       if (opcao < 0 || opcao > 3) {
         System.out.println ("Opcao deve ser valor inteiro " +
		"entre 0 e 3");
       } else {
         return opcao;
       }
     }
   }

   public static MembroUniv fabrica() {
      MembroUniv m = null;
      int opcaoI = MenuInclusao();
      switch (opcaoI) {
             case 0:
		   break;
             case 1: 
		   m = new Assalariado();
 	         break;
	       case 2: 
		   m = new Estudante();
 	         break;
	       case 3: 
		   m = new AssalEstud();
      }
      return m;
   }

   public MembroUniv() {
     matricula = ++matrAtual;
   }

   public MembroUniv (String n, Data a, Data i) {
     super (n, a);
     matricula = ++matrAtual;
     ingresso = new Data (i);
   }
   
   public MembroUniv (MembroUniv m) {
     super (m.nome, m.aniv);
     matricula = m.matricula;
     ingresso = new Data (m.ingresso);
   }
 
   public int retornaMatr() {
      return matricula;
   }

   public Data retornaIngresso() {
      return new Data(ingresso);
   }

   public String toString() {
        return super.toString() + 
        "\nMatricula: " + matricula + 
        "\nIngresso: " + ingresso;
   }

   public void leConsole() {
     super.leConsole();
     System.out.print("Ingresso: ");
     ingresso = new Data();
     ingresso.leConsole();
   }

   public boolean equals (Object o) {
     MembroUniv m = (MembroUniv) o;
     return matricula == m.matricula;
   } 


}

/* 
** Assalariado.java 
*/

interface AssalInter extends MUInter {
   float retornaSal();
   boolean maraja();
}

class Assalariado extends MembroUniv implements AssalInter {
   protected float salario;
   private static final float salMaraja = 8000.00f;

   public Assalariado(){};
  
   public Assalariado (String n, Data a, Data i, float s) {
     super (n, a, i);
     salario = s;
   }
   
   public Assalariado (Assalariado a) {
     super (a);
     salario = a.salario;
   }
 
   public float retornaSal() {
      return salario;
   }

   public String toString() {
     return super.toString() + "\nSalario: " + salario;
   }

   public boolean maraja() {
     if (salario >= salMaraja) {
        return true;
     } else {
        return false;
     }
   }

   public void leConsole() {
     super.leConsole();
     System.out.print("Salario: ");
     salario = Console.readFloat();
   }

}    

/* 
** Estudante.java 
*/

interface EstudInter extends MUInter {
   float retornaCR();
   boolean bomAluno();
} 

class Estudante extends MembroUniv implements EstudInter {
  protected float cr;
  private static float crBom = 7.0f;

  public Estudante() {}

  public Estudante (String n, Data a, Data i, float c) {
    super (n, a, i);
    cr = c;
  }
   
   public Estudante (Estudante a) {
    super (a);
    cr = a.cr;
  }
 
  public float retornaCR() {
      return cr;
  }

   public String toString() {
     return super.toString() + "\nCR: " + cr;
   }

   public boolean bomAluno() {
     if (cr >= crBom) {
        return true;
     } else {
        return false;
     }
   }

   public void leConsole() {
     super.leConsole();
     System.out.print("CR: ");
     cr = Console.readFloat();
   }

}    

/* 
** AssalEstud.java 
*/

class AssalEstud extends Assalariado implements AssalInter, EstudInter {
   Estudante e;     // heranca repetida de nome e aniv

   public AssalEstud() {
     e = new Estudante();
   }
  
   public AssalEstud (String n, Data a, Data ia, Data ie, float s, float c) {
     super (n, a, ia, s);
     e = new Estudante (n, a, ie, c);  
   }
   
   public AssalEstud (AssalEstud a) {
     super (a);
     e = new Estudante (a.e);
   }

   public float retornaCR() {
      return e.retornaCR();
   }

   public String toString() {
     return super.toString() +  "\nMatricula(aluno): " + e.retornaMatr()  
         + "\nIngresso(aluno): " + e.retornaIngresso() + "\nCR: " + 
         e.retornaCR();   
   }

   public boolean bomAluno() {
     return e.bomAluno();
   }

   public void leConsole() {
     super.leConsole();
     System.out.print ("Data de ingresso como estudante: ");
     e.ingresso = new Data();
     e.ingresso.leConsole();
     System.out.print ("CR: ");
     e.cr = Console.readFloat();
   }

}    

/* 
** Grupo.java 
*/

class Grupo {
  private Object [] p;
  private int tam = 0;
  private int marc = -1;

  Grupo () {
    p = new Object [5];  
  }

  Grupo (int n) {
    p = new Object [n];  
  }

  public void aumenta (int m) {
    Object [] novo = new Object [p.length + m];
    for (int i = 0; i < tam; i++) {
      novo[i] = p[i];
    }
    p = novo;
  }

  public void inclui (Object o) {
    if (tam == p.length) {
      aumenta (5);
    }
    p[tam++] = o;
  }

  public Object obtem (Object o) {
    for (int i = 0; i < tam; i++) {
      if (o.equals(p[i])) return p[i];
    }
    return null;
  }

  public Object obtem (int i) {
    if (i >= 0 && i < tam) return p[i];
    return null;
  }

  public void exclui (Object o) {
    for (int i = 0; i < tam;) {
      if (o.equals(p[i])) {
        for (int j = i; j < tam - 1; j++) {
          p[j] = p[j+1];
        }
        tam--;
	  p[tam] = null; 
      } else {
        i++;
      }
    }
  }

  public void exclui () {
     for (int j = marc; j < tam - 1; j++) {
        p[j] = p[j+1];
     }
     tam--;
     p[tam] = null; 
  }

  public int tamanho() {
    return tam;
  }
 
  public void inicio() {
    marc = -1;
  }

  public boolean fim() {
    return tam == 0 || marc == tam;
  }

  public Object proximo() {
    marc++;
    if (!fim()) {
      return p[marc];
    } else {
      inicio();
      return null;
    }
  }

  public String toString() {
    String s = "";
    for (int i = 0; i < tam; i++) {
      s += p[i] + "\n";
    }
    return s;
  }

  public void ordena (Comparator c, int ordem ) {
     /* 
        ordem negativo -> ordenacao decrescente
        ordem positivo -> ordenacao crescente
     */
        
     for (int j = 0; j < tam; j++) {
       for (int i = 0; i < tam-1; i++) {
          if (c.compare(p[i],p[i+1])*ordem > 0) {
            Object aux = p[i];
            p[i] = p[i+1];
            p[i+1] = aux;
          }
       }
     }
  }       

}

/* 
** Universidade.java 
*/

public class Universidade { 

   public static void main (String[] args) {

     System.out.print("Entre com o numero de pessoas: ");
     int n = Console.readInteger();
     Grupo gp = new Grupo(n);

     for (int i = 0; i < n; i++) {
           MembroUniv m = MembroUniv.fabrica();
           if (m != null) {
              m.leConsole();
              gp.inclui(m);
           }
     }

     gp.ordena(new Pessoa.ComparaNome(), 1);
     System.out.println(gp);
     Console.readChar();

     gp.ordena(new Pessoa.ComparaNome(), -1);
     System.out.println(gp);
     Console.readChar();

     gp.ordena(new Pessoa.ComparaAniv(), 1);
     System.out.println(gp);
     Console.readChar();

     gp.ordena(new Pessoa.ComparaAniv(), -1);
     System.out.println(gp);
     Console.readChar();

     Data hoje = new Data();
     hoje.leRelogio();
     for (int j=0; j<gp.tamanho(); j++) {
       PessoaInter p = (PessoaInter) gp.obtem(j);
       if(hoje.igualDiaMes(p.retornaAniv())){
         System.out.println (p.retornaNome() 
            + " e' aniversariante!");
       }

       if (p instanceof AssalInter) {
         AssalInter a = (AssalInter) p;
         if (a.maraja()) {
           System.out.println(a.retornaNome() + 
              " e' maraja'!");
         }
       }

       if (p instanceof EstudInter) {
         EstudInter a = (EstudInter) p;
         if (a.bomAluno()) {
           System.out.println(a.retornaNome() + 
              " e' bom Aluno!");
         }
       }
     }
   }
}    


